;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; File:         takl.sch
; Description:  TAKL benchmark from the Gabriel tests
; Author:       Richard Gabriel
; Created:      12-Apr-85
; Modified:     12-Apr-85 10:07:00 (Bob Shaw)
;               22-Jul-87 (Will Clinger)
; Language:     Scheme
; Status:       Public Domain
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; TAKL -- The TAKeuchi function using lists as counters.

(define (listn n)
  (if (zero? n)
      '()
      (cons n (listn (- n 1)))))

(define l18 (listn 18))
(define l12 (listn 12))
(define l6 (listn 6))

(define (mas x y z)
  (if (not-shorter? y x)
      z
      (mas (mas (cdr x) y z)
	   (mas (cdr y) z x)
	   (mas (cdr z) x y))))

(define (mas1 x y z)
  (if (shorter? y x)
      (mas1 (mas1 (cdr x) y z)
	   (mas1 (cdr y) z x)
	   (mas1 (cdr z) x y))
       z))

(define (takl x y z)
  (letrec ((shorter? 
	    (lambda (x y)
	      (cond ((null? y) #f)
		    ((null? x) #t)
		    (else (shorter? (cdr x) (cdr y))))))
	   (aux (lambda (x y z)
		  (if (shorter? y x)
		      (aux (aux (cdr x) y z)
			   (aux (cdr y) z x)
			   (aux (cdr z) x y))
		      z))))
    (aux x y z)))

;; Plus rapide que le suivant
(define (shorter1? x y)
  (cond ((null? y) #f)
	((null? x) #t)
	(else (shorter1? (cdr x) (cdr y)))))

(define (shorter? x y)
  (do ((x x (cdr x))
       (y y (cdr y)))
      ((or (null? x) (null? y)) 
       (not (null? y)))))

;; qui se traduit par :
;(lambda (x y) 
;  (letrec ((aux (lambda (x y) 
;		  (if (or (#'null? x) (#'null? y)) 
;		      (#'not (#'null? y)) 
;		      (aux (#'cdr x) (#'cdr y)))))) 
;    (aux x y)))

(define (not-shorter? x y)
  (>= (length x) (length y)))

;;; call: (mas l18 l12 l6) and (takl l18 l12 l6)

;; takl 14000 (15000) -- 2900 (3400)
;; mas1 14000 (15000) -- 3470 (3370)
;; mas  3600 (3200)   -- 2000 (550)
